# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import io
import os
import re
import shutil
import typing

from scss.compiler import compile_file  # type: ignore
import click

from aep_site.jinja.env import jinja_env
from aep_site.models.aep import AEP
from aep_site.models.scope import Scope
from aep_site.models.site import Site
from aep_site.models.page import Page


SUPPORT_ROOT = os.path.realpath(
    os.path.join(os.path.dirname(__file__), 'support'),
)


class Publisher:
    """Class for writing the AEP site out to disk."""

    def __init__(self, src, dest):
        self.output_dir = dest.rstrip(os.path.sep)
        self.site = Site.load(src.rstrip(os.path.sep))

    def log(self, message: str, *, err: bool = False):
        """Conditionally log a message."""
        click.secho(f'{message}', nl=True, err=err)

    def publish_site(self):
        """Publish the full site."""
        # Publish every page, AEP listing, and AEP.
        for page in self.site.pages.values():
            self.publish_doc(page)
        for scope in self.site.scopes.values():
            self.publish_doc(scope)
        for aep in self.site.aeps.values():
            self.publish_doc(aep)

        # Publish support pages that are not Markdown pages.
        self.publish_template('index.html.j2', 'index.html')
        self.publish_template('search.html.j2', 'search.html')

        # Copy over assets files verbatim.
        self.publish_assets()

        # Publish dynamic content that goes in the assets directory.
        self.publish_template(
            'search.js.j2',
            'assets/js/search/tipuesearch_content.js',
        )
        self.publish_css()

    def publish_doc(self, doc: typing.Union[AEP, Scope, Page]):
        """Publish a single AEP document to disk."""
        # Determine the path and make sure the directory exists.
        path = f'{self.output_dir}{doc.relative_uri}.html'
        os.makedirs(os.path.dirname(path), exist_ok=True)

        # Write the document's file to disk.
        with io.open(path, 'w') as f:
            f.write(doc.render())
        self.log(f'Successfully wrote {path}.')

        # If this document has any redirects (only AEPs do as of this writing),
        # write out the redirect files to disk also.
        for redirect in getattr(doc, 'redirects', set()):
            rpath = f'{self.output_dir}{redirect}.html'
            with io.open(rpath, 'w') as f:
                f.write(jinja_env.get_template('redirect.html.j2').render(
                    site=self.site,
                    target=doc,
                ))
            self.log(f'Successfully wrote {rpath} (redirect).')

    def publish_template(self, tmpl: str, path: str):
        """Publish a site-wide template to a particular path."""
        path = f'{self.output_dir}{os.path.sep}{path}'
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with io.open(path, 'w') as f:
            f.write(jinja_env.get_template(tmpl).render(
                site=self.site,
                path=path.split('.')[0],
            ))
        self.log(f'Successfully wrote {path}.')

    def publish_assets(self):
        """Copy the assets directory."""
        shutil.copytree(
            src=os.path.join(SUPPORT_ROOT, 'assets'),
            dst=os.path.join(self.output_dir, 'assets'),
            dirs_exist_ok=True,
        )
        self.log('Successfully wrote asset files.')

    def publish_css(self):
        """Compile and publish all SCSS files in the root SCSS directory."""
        scss_path = os.path.join(SUPPORT_ROOT, 'scss')
        css_path = os.path.join(self.output_dir, 'assets', 'css')
        os.makedirs(css_path, exist_ok=True)
        for scss_file in os.listdir(scss_path):
            if not scss_file.endswith('.scss'):
                continue
            css_file = os.path.join(
                css_path,
                re.sub(r'\.scss$', '.css', scss_file),
            )
            with io.open(css_file, 'w') as f:
                f.write(compile_file(os.path.join(scss_path, scss_file)))
            self.log(f'Successfully wrote {css_file}.')



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import markdown              # type: ignore
import pymdownx.highlight    # type: ignore
import pymdownx.superfences  # type: ignore

from aep_site.utils import cached_property


class MarkdownDocument(str):
    """A utility class representing Markdown content."""

    def __init__(self, content: str, *, toc_title: str = 'Contents'):
        self._content = content
        self._engine = markdown.Markdown(
            extensions=[_superfences, 'tables', 'toc', 'pymdownx.tabbed'],
            extension_configs={
                'toc': {
                    'title': toc_title,
                    'toc_depth': '2-3',
                },
            },
            tab_length=2,
        )

    def __str__(self) -> str:
        return self._content

    @cached_property
    def blocked_content(self) -> str:
        """Return a Markdown document with Jinja blocks added per-section."""
        answer = str(self)

        # We need to fire the html property because conversion may not have
        # happened yet, and the toc extension requires conversion to occur
        # first.
        #
        # The `html` property caches, so this does not entail any additional
        # performance hit.
        self.html

        # We use the Markdown toc plugin to separate the headers and identify
        # where to put the blocks. Iterate over each toc_token and encapsulate
        # it in a block.
        toc_tokens = self._engine.toc_tokens  # type: ignore
        for ix, token in enumerate(toc_tokens):
            try:
                next_token = toc_tokens[ix + 1]
            except IndexError:
                next_token = None
            answer = _add_block(answer, token, next_token)

        # Done; return the answer.
        return answer

    @cached_property
    def html(self) -> str:
        """Return the HTML content for this Markdown document."""
        return self._engine.convert(self._content)

    @cached_property
    def title(self) -> str:
        """Return the document title."""
        return self._content.strip().splitlines()[0].strip('# \n')

    @property
    def toc(self) -> str:
        """Return a table of contents for the Markdown document."""
        # We need to fire the html property because conversion may not have
        # happened yet, and the toc extension requires conversion to occur
        # first.
        #
        # The `html` property caches, so this does not entail any additional
        # performance hit.
        self.html
        return self._engine.toc  # type: ignore


def _add_block(content: str, toc_token, next_toc_token=None) -> str:
    # Determine the beginning of the block.
    heading = '#' * toc_token['level'] + r'\s+' + toc_token['name']
    match = re.search(heading, content)
    if not match:
        return content
    start_ix = match.span()[0]

    # Determine the end of the block.
    end_ix = len(content)
    if next_toc_token:
        end_h = '#' * next_toc_token['level'] + r'\s+' + next_toc_token['name']
        match = re.search(end_h, content)
        if not match:
            return content
        end_ix = match.span()[0]
    elif '{% endblock %}' in content[start_ix:]:
        # We can match against an {% endblock %} safely because we are
        # processing "top-down" (e.g. <h2> before <h3>), then beginning-to-end;
        # this means that encountering an {% endblock %} is guaranteed to be
        # the endblock for the enclosing block.
        end_ix = content.index('{% endblock %}', start_ix)

    # Add the block for this heading.
    block = content[start_ix:end_ix]
    block_id = toc_token['id'].replace('-', '_')
    content = content.replace(block, '\n'.join((
        f'{{% block {block_id} %}}',
        block,
        f'{{% endblock %}} {{# {block_id} #}}',
    )), 1)

    # Iterate over any children to this toc_token and process them.
    for ix, child in enumerate(toc_token['children']):
        try:
            next_child = toc_token['children'][ix + 1]
        except IndexError:
            next_child = None
        content = _add_block(content, child, next_toc_token=next_child)

    # Done; return the content with blocks.
    return content


def _fmt(src: str, lang: str, css_class: str, *args, **kwargs):
    """Custom Markdown formatter that retains language classes."""
    highlighter = pymdownx.highlight.Highlight(guess_lang=False)
    return highlighter.highlight(src, lang, f'{css_class} language-{lang}')


# Define the superfences extension.
# We need a custom fence to maintain useful language-specific CSS classes.
_superfences = pymdownx.superfences.SuperFencesCodeExtension(
    custom_fences=[{
        'name': '*',
        'class': 'highlight',
        'format': _fmt,
    }],
)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import re
import textwrap

import jinja2.ext
import jinja2.nodes


class SampleExtension(jinja2.ext.Extension):
    tags = {'sample'}

    def parse(self, parser):
        # The first token is the token that started the tag, which is always
        # "sample" because that is the only token we watch.
        lineno = next(parser.stream).lineno

        # Get the sample to be loaded.
        # A sample includes both a filename and a symbol within that file,
        # which is what is shown in the output.
        fn = parser.stream.expect('string').value
        symbols = []
        while parser.stream.look().type == 'string':
            parser.stream.expect('comma')
            symbols.append(parser.stream.expect('string').value)

        # Load the sample file.
        aep = self.environment.loader.aep
        filename = os.path.join(aep.path, fn)
        try:
            with io.open(filename, 'r') as f:
                code = f.read()
        except FileNotFoundError:
            raise jinja2.TemplateSyntaxError(
                filename=parser.filename,
                lineno=lineno,
                message=f'File not found: {filename}',
            )

        # Tease out the desired symbols and make snippets.
        snippets = []
        for symbol in symbols:
            match = re.search(rf'^([\s]*)({symbol})', code, flags=re.MULTILINE)
            if not match:
                raise jinja2.TemplateSyntaxError(
                    filename=parser.filename,
                    lineno=lineno,
                    message=f'Symbol not found: {symbol}',
                )

            # Determine the end of the symbol.
            # This attempts to parse C-style brace syntax if it encounters a
            # `{` character, or Python/YAML-style indenting if it encounters a
            # `:` character.
            #
            # The first thing we need to know is which syntax we are using.
            # We attempt to guess that by seeing which token we encounter next
            # after our symbol.
            start = match.start()
            try:
                ix, block_token = sorted([
                    (loc + 1, i) for i in (':', '{', ';')
                    if (loc := code.find(i, match.end())) != -1])[0]
            except IndexError:
                raise jinja2.TemplateSyntaxError(
                    filename=filename,
                    lineno=code.count('\n', 0, start) - 1,
                    message=f'No block character (:, {{) found after {symbol}',
                )

            # Push the start marker backwards to include any leading comments.
            lines = code[0:start - 1].split('\n')
            for line in reversed(lines):
                if re.search(r'^[\s]*(//|#)', line):
                    start -= len(line) + 1
                else:
                    break

            # If we got a `:`, we parse by indentation, stopping at the
            # start of the next line with the same indentation as our match.
            snippet = ''
            if block_token == ':':
                indent = len(match.groups()[0])
                end_match = re.search(r'^[\s]{0,%d}[\S]+' % indent, code[ix:],
                    re.MULTILINE,
                )
                if end_match:
                    snippet = code[start:end_match.start() + ix]
                else:
                    snippet = code[start:]
                snippet = textwrap.dedent(snippet)

            # We got a '{'; Find the corresponding closed brace.
            elif block_token == '{':
                cursor = match.start()
                while (close_brace := code.find('}', cursor)) != -1:
                    s, e = match.start(), close_brace + 1
                    if code.count('{', s, e) == code.count('}', s, e):
                        snippet = textwrap.dedent(code[start:e])
                        break
                    cursor = e
                else:
                    # Unable to find a corresponding closed brace; complain.
                    raise jinja2.TemplateSyntaxError(
                        filename=filename,
                        message=f'No corresponding }} found for {symbol}.',
                        lineno=code.count('\n', 0, start) - 1,
                    )

            # We got a ';'. Stop there.
            else:
                end = code.find(';', match.start()) + 1
                snippet = textwrap.dedent(code[start:end])

            # Append the snippet to the list of snippets.
            snippets.append(snippet)

        # We have a snippet. Time to put the Markdown together.
        md = '\n'.join((
            '```{0}'.format(filename.split('.')[-1]),
            '\n\n'.join(snippets),
            '```',
        ))

        # Finally, return a node to display it.
        return jinja2.nodes.Output(
            [jinja2.nodes.TemplateData(md)],
        )



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jinja2.ext
import jinja2.nodes


class TabExtension(jinja2.ext.Extension):
    tags = {'tab'}

    def parse(self, parser, cursor=()):
        # The first token is the token that started the tag, which is always
        # "tab" because that is the only token we watch.
        lineno = next(parser.stream).lineno

        # Get the language for this tab.
        lang = parser.stream.expect('name')
        lang_title = {
            'oas': 'OpenAPI 3.0',
            'proto': 'Protocol buffers',
        }.get(lang.value, lang.value.capitalize())

        # Encase the tab's content in a Markdown tab, properly indented.
        tab_title = jinja2.nodes.Output(
            [jinja2.nodes.TemplateData(f'=== "{lang_title}"\n\n')],
        )
        body = parser.parse_statements(['name:endtabs', 'name:tab'])
        indented_body = jinja2.nodes.FilterBlock(
            body,
            jinja2.nodes.Filter(
                None, 'indent', (), [
                    jinja2.nodes.Keyword('width', jinja2.nodes.Const(2)),
                    jinja2.nodes.Keyword('first', jinja2.nodes.Const(True)),
                ], None, None,
            ),
        ).set_lineno(lineno)
        cursor += (tab_title, indented_body)

        # If there is another tab, parse it too.
        if parser.stream.current.value == 'tab':
            return self.parse(parser, cursor=cursor)
        else:
            next(parser.stream)  # Drop endtabs.

        # Done; return the content.
        return list(cursor)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import jinja2
import jinja2.ext
import jinja2.nodes

from aep_site import md


TEMPLATE_DIR = os.path.realpath(
    os.path.join(os.path.dirname(__file__), '..', 'support', 'templates'),
)


# "Standard" jinja2 environment that loads from the filesystem.
jinja_env = jinja2.Environment(
    loader=jinja2.FileSystemLoader(searchpath=TEMPLATE_DIR),
    undefined=jinja2.StrictUndefined,
)
jinja_env.filters['markdown'] = lambda s: md.MarkdownDocument(s).html



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import dataclasses
import io
import re
import os
import typing

import yaml

from aep_site import md
from aep_site.jinja.env import jinja_env
from aep_site.utils import cached_property


@dataclasses.dataclass(frozen=True)
class Page:
    body: md.MarkdownDocument
    repo_path: str
    collection: Collection
    config: typing.Dict[str, typing.Any]

    @property
    def code(self) -> str:
        """Return this page's code.

        Note: Uniqueness is only guaranteed within the collection.
        """
        return self.repo_path.split('/')[-1].split('.')[0].lower()

    @cached_property
    def content(self) -> md.MarkdownDocument:
        """Return the Markdown content for this page."""
        # If this was originally a Jinja template, then we need to render
        # this specific content in isolation (apart from the page template).
        if self.repo_path.endswith('.j2'):
            return md.MarkdownDocument(
                jinja_env.from_string(self.body).render(site=self.site),
            )

        # This is not a template; just return the body directly.
        return self.body

    @property
    def relative_uri(self) -> str:
        return f'/{self.collection.code}/{self.code}'.replace('/general', '')

    @property
    def site(self) -> Site:
        return self.collection.site

    @property
    def title(self) -> str:
        """Return the page title."""
        return self.content.title

    def render(self) -> str:
        return jinja_env.get_template('page.html.j2').render(
            page=self,
            path=self.relative_uri,
            site=self.site,
        )


@dataclasses.dataclass(frozen=True)
class Collection:
    code: str
    site: Site

    @property
    def base_dir(self) -> str:
        return os.path.join(self.site.base_dir, 'pages', self.code)

    @cached_property
    def pages(self) -> typing.Dict[str, Page]:
        """Return the pages in this collection."""
        answer = {}

        # For the general collection, the CONTRIBUTING.md file is a special
        # case: we need it in the root so GitHub will find it.
        if self.code == 'general':
            answer['contributing'] = self._load_page(
                os.path.join(self.site.base_dir, 'CONTRIBUTING.md'),
            )

        # Iterate over the pages directory and load static pages.
        for fn in os.listdir(self.base_dir):
            # Sanity check: Ignore non-Markdown files.
            if not fn.endswith(('.md', '.md.j2')):
                continue

            # Load the page and add it to the pages dictionary.
            page = self._load_page(os.path.join(self.base_dir, fn))
            answer[page.code] = page
        return answer

    def _load_page(self, md_file: str) -> Page:
        """Load a support page and return a new Page object."""
        # Read the page file from disk.
        with io.open(md_file, 'r') as f:
            body = f.read()

        # Check for a config file. If one exists, load it too.
        config_file = re.sub(r'\.md(\.j2)?$', '.yaml', md_file)
        config = {}
        if os.path.exists(config_file):
            with io.open(config_file, 'r') as f:
                config = yaml.safe_load(f)

        # Return the page.
        return Page(
            body=md.MarkdownDocument(body),
            collection=self,
            config=config,
            repo_path=md_file[len(self.site.base_dir):],
        )


if typing.TYPE_CHECKING:
    from aep_site.models.site import Site



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re

from scss.compiler import compile_file  # type: ignore
import flask

from aep_site.jinja.env import jinja_env
from aep_site.models.site import Site


ROOT = os.path.realpath(os.path.join(os.path.dirname(__file__), '..'))
app = flask.Flask('aep', static_folder=f'{ROOT}/aep_site/support/assets/')
app.jinja_env = jinja_env  # type: ignore


@app.route('/')
def hello():
    return flask.render_template('index.html.j2', site=flask.g.site)


@app.route('/<int:aep_id>')
def aep(aep_id: int):
    """Display a single AEP document."""
    return flask.g.site.aeps[aep_id].render()


@app.route('/search')
def search():
    """Display the search page."""
    return flask.render_template('search.html.j2',
        site=flask.g.site,
        path=flask.request.path,
    )


@app.route('/<page>')
@app.route('/<collection>/<page>')
def page(page: str, collection: str = "general"):
    """Display a static page or listing of AEPs in a given scope."""
    site = flask.g.site
    if page in site.scopes:
        return site.scopes[page].render()
    if (
        collection in site.collections and
        page in site.collections[collection].pages
    ):
        return site.collections[collection].pages[page].render()
    if collection in site.scopes and int(page) in site.scopes[collection].aeps:
        return site.scopes[collection].aeps[int(page)].render()
    flask.abort(404)


@app.route('/assets/css/<path:css_file>')
def scss(css_file: str):
    """Compile the given SCSS file and return it."""
    scss_file = re.sub(r'\.css$', '.scss', css_file)
    css = compile_file(f'{ROOT}/aep_site/support/scss/{scss_file}')
    return flask.Response(css, mimetype='text/css')


@app.route('/assets/js/search/tipuesearch_content.js')
def search_content():
    """Compile the search content JavaScript and return it."""
    return flask.Response(
        flask.render_template('search.js.j2', site=flask.g.site),
        mimetype='text/javascript',
    )


def site_load_func(src: str):
    """Return a function that loads the site and registers it.

    This is used in cli.py to register to Flask.before_request.
    """
    def fx():
        flask.g.site = Site.load(src)

        # This is the dev server, so plow over whatever the configuration
        # says that the site URL is.
        flask.g.site.config.setdefault('urls', {})
        flask.g.site.config['urls']['site'] = 'http://localhost:4000'
    return fx



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import collections
import dataclasses
import datetime
import io
import re
import typing

import jinja2

from aep_site import md
from aep_site.jinja.env import jinja_env
from aep_site.jinja.ext.sample import SampleExtension
from aep_site.jinja.ext.tab import TabExtension
from aep_site.jinja import loaders
from aep_site.utils import cached_property


@dataclasses.dataclass(frozen=True)
class AEP:
    id: int
    state: str
    created: datetime.date
    scope: Scope
    path: str
    config: typing.Dict[str, typing.Any]
    changelog: typing.Set[Change] = dataclasses.field(default_factory=set)

    @cached_property
    def content(self) -> md.MarkdownDocument:
        # Render the content template into the actual content.
        answer = next(iter(self.templates.values())).render(
            aep=self,
            site=self.site,
        )

        # TEMPORARY: In the interest of having two discrete migrations,
        # rewrite the old link style to the new one.
        answer = re.sub(
            r'\.?\./0([\d]{1,3})\.md',
            lambda m: f'{self.site.relative_uri}/{int(m.groups()[0]):d}',
            answer,
        )

        # Hotlink AEP references.
        # We can (now) link to AEPs outside of general, those that are numbered
        # > 999, because the site redirects from 4221 to client-libraries/4221
        # now, for example.
        answer = re.sub(
            r'\b(?<!\[)AEP-(\d+)\b',  # AEP-###, but not after a `[`.
            fr'[AEP-\1]({self.site.relative_uri}/\1)',
            answer,
        )
        answer = re.sub(
            r'(?<=\])\[aep-(\d+)\]',  # [aep-###], after a `]`.
            fr'({self.site.relative_uri}/\1)',
            answer,
        )

        # Append the changelog if there is one.
        if self.changelog:
            answer += '\n\n## Changelog\n\n'
            for cl in sorted(self.changelog):
                answer += f'- **{cl.date}:** {cl.message}\n'

        # Return the document.
        return md.MarkdownDocument(answer)

    @cached_property
    def env(self) -> jinja2.Environment:
        return jinja2.Environment(
            extensions=[
                SampleExtension,
                TabExtension,
            ],
            loader=loaders.AEPLoader(self),
            undefined=jinja2.StrictUndefined,
        )

    @property
    def placement(self) -> Placement:
        """Return the placement data for this AEP, or sensible defaults."""
        p = self.config.get('placement', {})
        p.setdefault('category', self.scope.code)
        return Placement(**p)

    @property
    def redirects(self) -> typing.Set[str]:
        uris = {f'/{self.id:04d}', f'/{self.id:03d}', f'/{self.id:02d}'}
        uris = uris.difference({self.relative_uri})
        if 'redirect_from' in self.config:
            uris.add(self.config['redirect_from'])
        return uris

    @property
    def relative_uri(self) -> str:
        """Return the relative URI for this AEP."""
        if self.scope.code != 'general':
            return f'/{self.scope.code}/{self.id}'
        return f'/{self.id}'

    @property
    def repo_path(self) -> str:
        """Return the relative repository path."""
        return self.path[len(self.site.base_dir):]

    @property
    def site(self) -> Site:
        """Return the site for this AEP."""
        return self.scope.site

    @cached_property
    def templates(self) -> typing.Dict[str, jinja2.Template]:
        """Load and return the templates for this AEP."""
        # Sanity check: Is this a legacy AEP (in the old Jekyll format)?
        # If so, process out a single template and return it.
        if self._legacy:
            with io.open(self.path, 'r') as f:
                contents = f.read()
            _, body = contents.lstrip('-\n').split('---\n', maxsplit=1)
            return {'generic': jinja2.Template(body,
                undefined=jinja2.StrictUndefined,
            )}

        # Return a dictionary with all of the templates.
        #
        # Note: This could be made more efficient in the future by only loading
        # the template that the user wants.
        return collections.OrderedDict([
            (k, self.env.get_template(k)) for k in self.env.list_templates()
        ])

    @cached_property
    def title(self) -> str:
        """Return the AEP title."""
        return self.content.title

    @property
    def updated(self):
        """Return the most recent date on the changelog.

        If there is no changelog, return the created date instead.
        """
        if self.changelog:
            return sorted(self.changelog)[0].date
        return self.created

    @property
    def _legacy(self) -> bool:
        """Return True if this is a legacy AEP, False otherwise."""
        return self.path.endswith('.md')

    def render(self):
        """Return the fully-rendered page for this AEP."""
        return jinja_env.get_template('aep.html.j2').render(
            aep=self,
            path=self.relative_uri,
            site=self.site,
        )


@dataclasses.dataclass(frozen=True)
class Change:
    date: datetime.date
    message: str

    def __hash__(self):
        return hash(str(self))

    def __lt__(self, other: 'Change'):
        # We sort changes in reverse-cron, so a later date is "less" here.
        return self.date > other.date

    def __str__(self):
        return f'{self.date.isoformat()}: {self.message}'


@dataclasses.dataclass(frozen=True)
class Placement:
    category: str
    order: typing.Union[int, float] = float('inf')


if typing.TYPE_CHECKING:
    from aep_site.models.scope import Scope
    from aep_site.models.site import Site



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import click

from aep_site import server
from aep_site.publisher import Publisher


@click.command()
@click.argument('src', type=click.Path(
    allow_dash=False,
    dir_okay=True,
    exists=True,
    file_okay=False,))
@click.argument('dest', type=click.Path(
    allow_dash=False,
    dir_okay=True,
    exists=False,
    file_okay=False,
    writable=True))
def publish(src: str, dest: str):
    """Publish static HTML pages into the specified location."""
    publisher = Publisher(src, dest)
    publisher.publish_site()


@click.command()
@click.argument('src', type=click.Path(
    allow_dash=False,
    dir_okay=True,
    exists=True,
    file_okay=False))
def serve(src: str):
    """Run a debug server."""
    server.app.before_request(server.site_load_func(src))
    server.app.run(host='0.0.0.0', port=4000, debug=True)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import re
import typing

import jinja2
import jinja2.loaders

from aep_site.md import MarkdownDocument

if typing.TYPE_CHECKING:
    from aep_site.models.aep import AEP


class AEPLoader(jinja2.loaders.BaseLoader):
    """Loader for loading AEPs."""

    def __init__(self, aep: 'AEP'):
        super().__init__()
        self.aep = aep

    def get_source(self, env, template: str) -> typing.Tuple[str, str, None]:
        # Find the appropriate AEP file.
        if template == 'generic':
            fn = os.path.join(self.aep.path, 'aep.md')
        else:
            view = template.split('.')
            while view:
                view_str = '.'.join(view)
                fn = os.path.join(self.aep.path, f'aep.{view_str}.md')
                if os.path.isfile(fn) or os.path.isfile(f'{fn}.j2'):
                    break
                view = view[1:]

        # Sanity check: Does the file exist?
        # If not, raise an error.
        if not os.path.isfile(fn) and not os.path.isfile(f'{fn}.j2'):
            raise jinja2.TemplateNotFound(
                f'Could not find {template} template for AEP-{self.aep.id}.',
            )

        # Are we loading a plain file or a Jinja2 template file?
        if os.path.isfile(f'{fn}.j2'):
            fn += '.j2'

        # Load the contents.
        with io.open(fn) as f:
            contents = f.read()

        # Add custom {% block %} tags corresponding to Markdown headings.
        #
        # Note: We only do this if the template does not already have
        # {% block %} tags to avoid either stomping over input, or creating
        # an invalid template.
        if not re.search(r'\{%-? block', contents):
            # Iterate over the individual components in the table
            # of contents and make each into a block.
            contents = MarkdownDocument(contents).blocked_content

        # Return the template information.
        return contents, fn, None

    def list_templates(self) -> typing.List[str]:
        answer = []

        # We sort the files in the directory to read more specific
        # files first.
        exts_regex = r'(\.(j2|md|proto|oas|yaml))*$'
        for fn in sorted(os.listdir(self.aep.path),
                key=lambda p: len(re.sub(exts_regex, '', p).split('.')),
                reverse=True):

            # Each file may specify a view, which corresponds to a separate
            # template.
            view = '.'.join(re.sub(exts_regex, '', fn).split('.')[1:])
            if view and view not in answer:
                answer.append(view)

        # There is always a generic view.
        answer.append('generic')
        return answer



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import collections
import dataclasses
import io
import os
import typing

import yaml

from aep_site.jinja.env import jinja_env
from aep_site.models.aep import AEP
from aep_site.models.aep import Change
from aep_site.utils import cached_property


@dataclasses.dataclass(frozen=True)
class Scope:
    base_dir: str
    code: str
    title: str
    order: int
    site: Site
    config: typing.Dict[str, typing.Any]

    @cached_property
    def aeps(self) -> typing.Dict[int, AEP]:
        """Return a dict of AEPs, sorted by number."""
        answer = []

        # Load each AEP under this scope.
        # Similar to scopes, AEPs are detected based solely on
        # presence of file on disk.
        for fn in os.listdir(self.base_dir):
            path = os.path.join(self.base_dir, fn)
            aep_config = os.path.join(path, 'aep.yaml')
            if os.path.isdir(path) and os.path.exists(aep_config):
                with io.open(aep_config, 'r') as f:
                    meta = yaml.safe_load(f)

            # Sanity check: Does this look like an old-style AEP?
            elif path.endswith('.md'):
                # Load the AEP.
                with io.open(path, 'r') as f:
                    contents = f.read()

                # Parse out the front matter from the Markdown content.
                fm, body = contents.lstrip('-\n').split('---\n', maxsplit=1)
                meta = yaml.safe_load(fm)

            # This is not a recognized AEP.
            else:
                meta = None  # Not needed; fixes a coverage bug.
                continue

            # Create the AEP object.
            answer.append(AEP(
                id=meta.pop('id'),
                config=meta,
                changelog={Change(**i) for i in meta.pop('changelog', [])},
                created=meta.pop('created'),
                path=path,
                scope=self,
                state=meta.pop('state'),
            ))

        answer = sorted(answer, key=lambda i: i.id)
        return collections.OrderedDict([(i.id, i) for i in answer])

    @cached_property
    def categories(self) -> typing.Dict[str, Category]:
        # There may be no categories. This is fine; just "fake" a single
        # category based on the scope.
        #
        # This is guaranteed to have all AEPs, so just return immediately
        # and skip remaining processing.
        if not self.config.get('categories', None):
            return {self.code: Category(
                code=self.code,
                title=self.title,
                order=0,
                aeps=self.aeps,
            )}

        # Iterate over each category and piece together the AEPs in order.
        cats = collections.OrderedDict()
        for ix, cat in enumerate(self.config['categories']):
            # Determine what AEPs are in this category.
            aeps: typing.List[AEP] = []
            for aep in self.aeps.values():
                if aep.placement.category != cat['code']:
                    continue
                aeps.append(aep)
            aeps = sorted(aeps, key=lambda i: i.placement.order)

            # Create the category object and add it.
            cats[cat['code']] = Category(
                code=cat['code'],
                title=cat.get('title', cat['code'].capitalize()),
                order=ix,
                aeps=collections.OrderedDict([(i.id, i) for i in aeps]),
            )
        return cats

    @property
    def relative_uri(self) -> str:
        return f'/{self.code}'

    def render(self):
        """Render the page for this scope."""
        return jinja_env.get_template('aep-listing.html.j2').render(
            path=self.relative_uri,
            scope=self,
            site=self.site,
        )


@dataclasses.dataclass(frozen=True)
class Category:
    code: str
    title: str
    order: int
    aeps: typing.Dict[int, AEP]


if typing.TYPE_CHECKING:
    from aep_site.models.site import Site



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os

from setuptools import find_packages, setup  # type: ignore


PACKAGE_ROOT = os.path.abspath(os.path.dirname(__file__))

with io.open(os.path.join(PACKAGE_ROOT, 'VERSION'), 'r') as version_file:
    VERSION = version_file.read().strip()

with io.open(os.path.join(PACKAGE_ROOT, 'README.md'), 'r') as readme_file:
    long_description = readme_file.read().strip()

setup(
    name='aep-site-generator',
    version=VERSION,
    license='Apache 2.0',
    author='Luke Sneeringer',
    author_email='lukesneeringer@google.com',
    url='https://github.com/aep-dev/site-generator.git',
    packages=find_packages(exclude=['tests']),
    description='Static site generator for aep.dev and forks.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    entry_points="""[console_scripts]
        aep-site-gen=aep_site.cli:publish
        aep-site-serve=aep_site.cli:serve
    """,
    platforms='Posix; MacOS X',
    include_package_data=True,
    install_requires=(
        'click==8.1.3',
        'flask==2.2.2',
        'itsdangerous==2.1.2',
        'jinja2==3.1.2',
        'markdown==3.4.1',
        'markupsafe==2.1.1',
        'pygments==2.13.0',
        'pymdown-extensions==9.7',
        'pyscss==1.4.0',
        'pyyaml==6.0.1',
        'six==1.16.0',
        'types-Markdown==3.4.2.1',
        'types-PyYAML==6.0.12',
        'werkzeug==2.2.2',
    ),
    python_requires='>=3.8',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Code Generators',
    ],
    zip_safe=False,
)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import collections
import dataclasses
import io
import os
import typing
import uuid

import yaml

from aep_site.models.aep import AEP
from aep_site.models.page import Collection
from aep_site.models.page import Page
from aep_site.models.scope import Scope
from aep_site.utils import cached_property


@dataclasses.dataclass(frozen=True)
class Site:
    base_dir: str
    revision: str
    config: typing.Dict[str, typing.Any]

    @classmethod
    def load(cls, base_dir: str) -> Site:
        # Load all site configuration.
        config: typing.Dict[str, typing.Any] = {}
        config_dir = os.path.join(base_dir, 'config')
        for fn in os.listdir(config_dir):
            with io.open(os.path.join(config_dir, fn), 'r') as f:
                config[fn[:-5]] = yaml.safe_load(f)

        # Return a new Site object.
        return cls(
            base_dir=base_dir,
            config=config,
            revision=str(uuid.uuid4())[-8:],
        )

    @cached_property
    def aeps(self) -> typing.Dict[int, AEP]:
        """Return all the AEPs in the site."""
        answer = collections.OrderedDict()
        for scope in self.scopes.values():
            for id, aep in scope.aeps.items():
                answer[id] = aep
        return answer

    @property
    def base_url(self) -> str:
        """Return the site's base URL."""
        return self.config.get('urls', {}).get('site', '').rstrip('/')

    @cached_property
    def collections(self) -> typing.Dict[str, Collection]:
        """Return all of the page collections in the site."""
        pages_dir = os.path.join(self.base_dir, 'pages')
        answer = {}
        for dirname in os.listdir(pages_dir):
            # Sanity check: Is this a directory?
            if not os.path.isdir(os.path.join(pages_dir, dirname)):
                continue
            answer[dirname] = Collection(code=dirname, site=self)
        return answer

    @property
    def pages(self) -> typing.Dict[str, Page]:
        """Return all of the static pages in the site."""
        answer = {}
        for col in self.collections.values():
            for page in col.pages.values():
                # Disambiguate the pages from one another, but treat the
                # general collection as special and remove the prefix.
                code = f'{col.code}/{page.code}'.replace('general/', '')
                answer[code] = page
        return answer

    @property
    def relative_uri(self) -> str:
        return '/{}'.format('/'.join(self.base_url.split('/')[3:])).rstrip('/')

    @property
    def repo_url(self) -> str:
        return self.config.get('urls', {}).get('repo', '').rstrip('/')

    @cached_property
    def scopes(self) -> typing.Dict[str, Scope]:
        """Return all of the AEP scopes present in the site."""
        answer_list = []
        aep_dir = os.path.join(self.base_dir, 'aep')
        for fn in os.listdir(aep_dir):
            # If there is a scope.yaml file, then this is a scope directory
            # and we add it to our list.
            scope_file = os.path.join(aep_dir, fn, 'scope.yaml')
            if not os.path.exists(scope_file):
                continue

            # Open the scope's configuration file and create a Scope
            # object based on it.
            with io.open(scope_file, 'r') as f:
                conf = yaml.safe_load(f)
            code = conf.pop('code', fn)
            scope = Scope(
                base_dir=os.path.join(aep_dir, fn),
                code=code,
                config=conf,
                order=conf.pop('order', float('inf')),
                site=self,
                title=conf.pop('title', code.capitalize()),
            )

            # Append the scope to our list.
            answer_list.append(scope)

        # Create an ordered dictionary of scopes.
        answer = collections.OrderedDict()
        for scope in sorted(answer_list, key=lambda i: i.order):
            answer[scope.code] = scope
        return answer
